diff mbox series

[SRU,Jammy,3/4] mctp: make __mctp_dev_get() take a refcount hold

Message ID 20230704142113.581071-4-cascardo@canonical.com
State New
Headers show
Series CVE-2023-3439 | expand

Commit Message

Thadeu Lima de Souza Cascardo July 4, 2023, 2:21 p.m. UTC
From: Matt Johnston <matt@codeconstruct.com.au>

Previously there was a race that could allow the mctp_dev refcount
to hit zero:

rcu_read_lock();
mdev = __mctp_dev_get(dev);
// mctp_unregister() happens here, mdev->refs hits zero
mctp_dev_hold(dev);
rcu_read_unlock();

Now we make __mctp_dev_get() take the hold itself. It is safe to test
against the zero refcount because __mctp_dev_get() is called holding
rcu_read_lock and mctp_dev uses kfree_rcu().

Reported-by: Jakub Kicinski <kuba@kernel.org>
Signed-off-by: Matt Johnston <matt@codeconstruct.com.au>
Signed-off-by: David S. Miller <davem@davemloft.net>
(backported from commit dc121c0084910db985cf1c8ba6fce5d8c307cc02)
[cascardo: no test, no extended addr support,
 small conflict at mctp_pkttype_receive]
CVE-2023-3439
Signed-off-by: Thadeu Lima de Souza Cascardo <cascardo@canonical.com>
---
 net/mctp/device.c | 21 ++++++++++++++++++---
 net/mctp/route.c  |  2 ++
 2 files changed, 20 insertions(+), 3 deletions(-)
diff --git a/net/mctp/route.c b/net/mctp/route.c
index 1a60fee0e99e..9cd3ab18faca 100644
--- a/net/mctp/route.c
+++ b/net/mctp/route.c
@@ -860,11 +860,13 @@  static int mctp_pkttype_receive(struct sk_buff *skb, struct net_device *dev,
 		goto err_drop;
 
 	mctp_do_route(rt, skb);
+	mctp_dev_put(mdev);
 
 	return NET_RX_SUCCESS;
 
 err_drop:
 	kfree_skb(skb);
+	mctp_dev_put(mdev);
 	return NET_RX_DROP;
 }
diff mbox series

Patch

diff --git a/net/mctp/device.c b/net/mctp/device.c
index f556c6d01abc..7904eb599904 100644
--- a/net/mctp/device.c
+++ b/net/mctp/device.c
@@ -24,12 +24,25 @@  struct mctp_dump_cb {
 	size_t a_idx;
 };
 
 /* unlocked: caller must hold rcu_read_lock */
 struct mctp_dev *__mctp_dev_get(const struct net_device *dev)
 {
-	return rcu_dereference(dev->mctp_ptr);
+	struct mctp_dev *mdev = rcu_dereference(dev->mctp_ptr);
+
+	/* RCU guarantees that any mdev is still live.
+	 * Zero refcount implies a pending free, return NULL.
+	 */
+	if (mdev)
+		if (!refcount_inc_not_zero(&mdev->refs))
+			return NULL;
+	return mdev;
 }
 
+/* Returned mctp_dev does not have refcount incremented. The returned pointer
+ * remains live while rtnl_lock is held, as that prevents mctp_unregister()
+ */
 struct mctp_dev *mctp_dev_get_rtnl(const struct net_device *dev)
 {
 	return rtnl_dereference(dev->mctp_ptr);
@@ -109,6 +122,7 @@  static int mctp_dump_addrinfo(struct sk_buff *skb, struct netlink_callback *cb)
 				if (mdev) {
 					rc = mctp_dump_dev_addrinfo(mdev,
 								    skb, cb);
+					mctp_dev_put(mdev);
 					// Error indicates full buffer, this
 					// callback will get retried.
 					if (rc < 0)
@@ -254,7 +268,7 @@  void mctp_dev_hold(struct mctp_dev *mdev)
 
 void mctp_dev_put(struct mctp_dev *mdev)
 {
-	if (refcount_dec_and_test(&mdev->refs)) {
+	if (mdev && refcount_dec_and_test(&mdev->refs)) {
 		dev_put(mdev->dev);
 		kfree_rcu(mdev, rcu);
 	}
@@ -308,6 +322,7 @@  static size_t mctp_get_link_af_size(const struct net_device *dev,
 	if (!mdev)
 		return 0;
 	ret = nla_total_size(4); /* IFLA_MCTP_NET */
+	mctp_dev_put(mdev);
 	return ret;
 }
 
@@ -24,12 +24,25 @@  struct mctp_dump_cb {
 };
 
-/* unlocked: caller must hold rcu_read_lock */
+/* unlocked: caller must hold rcu_read_lock.
+ * Returned mctp_dev has its refcount incremented, or NULL if unset.
+ */
 struct mctp_dev *__mctp_dev_get(const struct net_device *dev)
 {